/*
THIS CODE USES THE GSL LIBRARY
Must link using libgsl.so
*/

#include <stdlib.h>
#include <stdio.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <string.h>
#include <math.h>
#include <sys/time.h>
#include <time.h>
#include <errno.h>

#define ASTRUCT_MAXROWS 1
#define ASTRUCT_MAXCOLS 1

extern double Mygsl_rhy_dists(int distribnumber, int counter, const gsl_rng * r);

struct the2dstruct
{
    int structnum;
    int nrows;
    int ncols;
    float a2dstruct[ASTRUCT_MAXROWS][ASTRUCT_MAXCOLS];
};
//------------- I N F O ------------------
//function takes in a two dimensional array,
//and number of pairs to be returned
//
//2d array pairs are each loaded into a struct
//ex. (0,0) and (0,1)
//
//then an array of those structs is created
//then randomized, returning a 2 dim array
//of the number of pairs
//------------- E N D  I N F O ------------------
float *random2dstructs_gsl(int inrows, int incols, float **input2darray, int returnrows, int returncols, int rhydistrib)
{
    char c;
    //mem ptrs for ran gen
    const gsl_rng_type * T;
    gsl_rng * r;

    int spinrows = returnrows;
    int spincols = returncols;
    int spincount = returnrows * returncols;
    int hh = inrows * incols;
    int h; //highest number
    int l = 1; //lowest number
    int f, i, j, k, nn, x, y, oddnumrange;
    int rhydist = rhydistrib;
    int distcounter;

    float **s = NULL;  //the local 2d array
    float *sPool, *curS_Ptr; //for calloc
    float *ss = NULL;//local 1d arrays
    float minpdf;
    int *pp = NULL;
    int *tt = NULL;

    double u; //value returned from Mygsl_rhy_dists

    struct the2dstruct **Arrayof2dStructs = NULL;
    struct timeval  t1; //structs for rand using time

first:

    distcounter = 1;  //sets or resets counter
//*****************************************
//********  ALLOCATE MEMORY ***************
//*****************************************

    /* create a generator chosen by the
       environment variable GSL_RNG_TYPE.
       In this case using 'taus' algo */
    gsl_rng_env_setup();
    T = gsl_rng_taus; //use taus algo as ran gen
    r = gsl_rng_alloc (T);  //alloc mem, see free below

//allocate local initialized 2d array memory
//NOTE: use of ptr pool for ease of freeing arrays later in the func
    s=(float **) calloc(spinrows+1, sizeof(float*));
    sPool = (float*)calloc(spincols*spinrows,sizeof(float));
    curS_Ptr=sPool;
    for (i = 0; i<spinrows+1; i++)
    {
        s[i]= curS_Ptr;
        curS_Ptr += spincols;
    }

//allocate local initialized 1d array memory
    ss=(float *) calloc(spincount, sizeof(float*));
    pp=(int *) calloc(spincount, sizeof(int*));
    tt=(int *) calloc(spincount, sizeof(int*));

//allocate memory for 1 2d struct
    Arrayof2dStructs = (struct the2dstruct *)calloc(1, sizeof(struct the2dstruct));
//allocate memory for the inrows*incols num of 2dstructs
    h= inrows*incols;
    for (x=0; x<=h; x++)
    {
        /* add ONE element to the array */
        Arrayof2dStructs= (struct the2dstruct **)realloc(Arrayof2dStructs, (x + 1) * sizeof(struct the2dstruct *));
        /* allocate memory for x number of the2dstructs */
        Arrayof2dStructs[x] = (struct the2dstruct *)malloc(sizeof(struct the2dstruct));
    }

//load data from incoming 2d array to local array of structs
//structnums begin with '1' for ease of counting and printing,
//but structs begin with '0'
    x=0;
    y=1;
    for (j = 0; j<inrows; j++)
    {
        for (k = 0; k<incols; k++)
        {
            Arrayof2dStructs[x]->a2dstruct[0][0] = input2darray[j][k];
            Arrayof2dStructs[y]->structnum = y;
            x++;
            y++;
        }
    }

//load only the odd numbers into a local 1d array
    oddnumrange=1;
    for (x=1; x<=hh; x++)
    {
        if(x%2==1)
        {
            pp[oddnumrange] = x;
            oddnumrange++;
        }
    }

    h=oddnumrange;  //highest equals the number of values + 1 (from oddnumrange++ above)

    for (x=1; x<=spincount/2; x++)
    {
next:
        gettimeofday(&t1, NULL);
        gsl_rng_default_seed = t1.tv_usec * t1.tv_sec; //use time as gen seed

        //****GSL. Get a random distribution****
        u = Mygsl_rhy_dists(rhydist, distcounter, r);

        f=(int)(u*h-1)+1; //limits between ranges and

        if (distcounter > 3000)
        {
            free (Arrayof2dStructs);
            free(*s);
            free(s);
            free(ss);
            free(pp);
            free(tt);
            gsl_rng_free (r);
            printf("\n The distribution is not generating numbers\
                \n within the ranges specified.  Try entering\
                \n different input values for the distribution. \n");
            goto first;
        }

        if(h==1)   //avoids rare case where highest = 1
        {
            if(f<l || f>h)
            {
                distcounter++;
                goto next;
            }
        }
        else
        {
            if(f<l || f>h-1)  /*avoid out of bounds nums */
            {
                distcounter++;
                goto next;
            }
        }
        tt[x]= pp[f];
        distcounter++;

    } //end for loop

// Fill odd slots
// uses the randomized numbers, from above
    x=0;
    float myfloat1;
    for(i=1; i<=spincount/2; i++)
    {
        nn=tt[i];
        x=x+1;
        ss[x]= Arrayof2dStructs[nn-1]->a2dstruct[0][0];
        x++;
    }

// Fill even slots with the even pair
    x=2;
    float myfloat2;
    for(i=1; i<=spincount/2; i++)
    {
        nn=tt[i];
        ss[x]=Arrayof2dStructs[nn]->a2dstruct[0][0];
        x=x+2;
    }

// loads the 1d array into a 2d array to return to the
// rhy2Dinput_gsl func below
    printf("\n...printing the spincount of 2d spun structs (from rand2dstructs_gsl)...\n");
    i=0;
    for (j = 0; j<spinrows; j++)
    {
        for (k = 0; k<spincols; k++)
        {
            i++;
            s[j][k] = ss[i];
            printf(" %4.6f ", s[j][k]);
        }
    }

//switch here to do another spin before returning values
    printf("\ntry again? (y)es, (n)o: ");
    while (getchar()!='\n');
    scanf("%c", &c);
    switch(c)
    {
    case 'y':
        //free ALL memory allocated so far
        for (x=0; x<=h; x++)
        {
            free(Arrayof2dStructs[x]);
        }
        free (Arrayof2dStructs);
        free(*s);//NOTE: this how ptr pool helps with freeing memory
        free(s);
        free(ss);
        free(pp);
        free(tt);
        gsl_rng_free (r);
        goto first;
        break;
    case 'n':
        printf("\nbye");
        break;
    }//ends switch
    printf("\n");

    for (x=0; x<=h; x++)
    {
        free(Arrayof2dStructs[x]);
    }
    free (Arrayof2dStructs);
    free(ss);
    free(pp);
    free(tt);
    gsl_rng_free (r);

    return s; //return only the local 2d array of randomized nums
    //not the whole struct to rhy2Dinput_gsl, below

} //end random2dstructs_gsl

//++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++

//******   rhy2Dinput_gsl    ******

//++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++
//++++++++++++++++++++++++++++++++

float *rhy2Dinput_gsl(float *rhy2Dreturnarray)
{
    char c;
    float **struct2dinputarray = NULL; //local 2d arrays
    float **struct2dreturnarray = NULL;
    float  *struct1dreturnarray = NULL;//local 1d array
    int i, j, k, nn, x;
    int rows, spinrows;  //number of input pairs and returned pairs
    int cols = 2;  //one input pair (for 2d leave cols as "2")
    int spincols = 2; //one returned pair
    float *inputPool, *returnPool; //for calloc
    float *curInputPtr, *curReturnPtr; //for calloc

    int rhydist;//the distribution selection

//**************************************
//********  ENTER VALUES ***************
//**************************************

//fill a 2d array to send to random2dstructs_gsl function, above
start:
    printf("\n");
    printf("Enter the number of pairs: ");
    scanf("%d",&rows);

//allocate 2d array memory
//keep 2d array as 2 cols
//But rows * num of pairs are needed for input
//NOTE: use of ptr pool for ease of freeing arrays later in the func
    struct2dinputarray=(float **) calloc(rows, sizeof(float*));
    inputPool = (float*)calloc(cols*rows,sizeof(float));
    curInputPtr=inputPool;
    for (i = 0; i<rows; i++)
    {
        struct2dinputarray[i]= curInputPtr;
        curInputPtr += cols;
    }
//load the 2d array with rhythmic pairs, ex:
//(0,0)=1.5, (0,,1)=0.5
//(1,0)=0.75, (1,1)=0.25, etc.
    x=1;
    for (j = 0; j<rows; j++)
    {
        for (k = 0; k<cols; k++)
        {
            printf("Enter value #%d: ", x);
            scanf("%f",&struct2dinputarray [j][k]);
            x++;
        }
    }
//------------- I N F O ------------------
// todo:  add error checking here to make sure each pair
// sums to a whole number.  In that way "Beats" returned
// will be clear.
// However could also just leave it up to the user to manage the
// the entry of values, so for example syncopation could result
//------------- E N D  I N F O ------------------
    printf("\n...here's the list \n");
    for (j = 0; j<rows; j++)
    {
        for (k = 0; k<cols; k++)
        {
            printf("%2.4f\n",  struct2dinputarray[j][k]);
        }
    }

//switch for another spin before returning values

    printf("try again? (y)es, (n)o: ");
    while (getchar()!='\n');
    scanf("%c", &c);
    switch(c)
    {
    case 'y':
//free 2d array memory
//NOTE: this how ptr pool helps with freeing memory
        free(*struct2dinputarray);
        free(struct2dinputarray);
        goto start;
        break;
    case 'n':
        goto next;
    break;
        default:
        break;
    }//ends switch
    printf("\n");
//***************************************
//********  RETURN VALUES ***************
//***************************************
next:
//number of pairs to be returned
    printf("\nEnter the number of pairs to be returned: ");
    scanf("%d",&spinrows);

//allocate 1d array memory
    struct1dreturnarray=(float **) calloc((spinrows*spincols), sizeof(float*));

//allocate 2d array memory
//NOTE: use of ptr pool for ease of freeing arrays later in the func
    struct2dreturnarray=(float **) calloc(spinrows, sizeof(float*));
    returnPool = (float*)calloc(spincols*spinrows,sizeof(float));
    curReturnPtr = returnPool;
    for (i = 0; i<spinrows; i++)
    {
        struct2dreturnarray[i]= curReturnPtr;
        curReturnPtr += spincols;
    }

//***************************************
//*****  SELECT RANDOM DISTRIBUTION *****
//***************************************

    printf("\n");
    printf("\t\t\t***** RHYTHM DISTRIBUTIONS *****");
    printf("\n1 = gaussian, 2 = gaussian_tail, 3 = exponential, 4 = laplace\
            \n5 = exppow, 6 = cauchy, 7 = rayleigh, 8 = rayleigh_tail\
            \n9 = landau, 10 = levy, 11 = levy_skew, 12 = gamma\
            \n13 = flat, 14 = lognormal, 15 = chisq, 16 = fdist\
            \n17 = tdist, 18 = beta, 19 = logistic, 20 = pareto\
            \n21 = Weibell, 22 = gumbel1, 23 = gumbel2, 24 = poisson\
            \n25 = binomial, 26 = negative_binomial, 27 = pascal\
            \n28 = geometric, 29 = hypergeometric, 30 = logarithmic\
            \n\nEnter a distribution for rhythm:  ");
    scanf("%d",&rhydist);

//***************************************
//********  CALL random2dstructs_gsl ********
//***************************************
    struct2dreturnarray = random2dstructs_gsl(rows, cols, struct2dinputarray, spinrows, spincols, rhydist);

// convert the 2d array from random2dstructs_gsl function to a 1d array
// to RETURN to main, for use with Cscore
    x=0;
    for (j = 0; j<spinrows; j++)
    {
        for (k = 0; k<spincols; k++)
        {
            x++;
            struct1dreturnarray[x] = struct2dreturnarray[j][k];
        }
    }

//free 2d array memory
//NOTE: this how ptr pool helps with freeing memory
    free(*struct2dinputarray);
    free(struct2dinputarray);

    free(*struct2dreturnarray);
    free(struct2dreturnarray);

    return struct1dreturnarray;   //returns a 1d array of rhythms

    free(struct1dreturnarray);//this does no good here

} //end rhy2Dinput_gsl
